#include "civcheat.h"

/************************************************************************/
/* Global Variable List:                                                */
/* UNSIGNED CHAR                                                        */
/* ---------------                                                      */
/* inputname[]      Holds the path & name of the CIV save file          */
/* memarea          Holds the image of the Civilization save file       */
/*                                                                      */
/* INTs                                                                 */
/* ---------------                                                      */
/* CIVSVE_handle    Handle for CIVIL?.SVE file specified by inputname[] */
/*                                                                      */
/* UNSIGNED INTs                                                        */
/* ---------------                                                      */
/* civ_num          Trailing number of the CIV save file                */
/* culture_number   Stored at the third byte of the file, this number   */
/*                  times 2 is the relative offset to be added to 312   */
/*                  to find where your money bytes are located.         */
/*                                                                      */
/* ERROR LEVELS:                                                        */
/* ---------------                                                      */
/*  0               No errors; successfully modified and backed up file */
/*  1               Error in command line; incorrect switch or no file  */
/*                     name specified with file name switch             */
/* 10               Error reading/writing to file                       */
/* 20               Memory/resource allocation error - very bad         */
/************************************************************************/

unsigned char   inputname[128]="CIVILX.SVE",*memarea;

int             CIVSVE_handle=0;

unsigned int    civ_num=0;

/************************************************************************/
/* parse_command_line():Figures out argv parms and sets appropriate var-*/
/*                      iables like inputname and exepath.              */
/************************************************************************/

void parse_command_line(int argc,char *argv[])
{
    unsigned char   c;                          /* Generic holding character */

    while (--argc > 0 )
    {
        if ( strlen( *++argv ) > 1 )
            c = (*argv)[1];
        else
            c = (*argv)[0];
        c=tolower( c );
        switch( c )
        {
            case 'w':
                terminate_program( WHATIS,1 );
                break;
            case '?':
                terminate_program( USAGETEXT,1 );
                break;
            case 'i':
                if( --argc > 0 )
                    strcpy( inputname, *++argv );
                else
                    terminate_program("Requires a filename parameter.\n",1);
                break;
            default:
                if ( c<48 && c>57 )
                    terminate_program( USAGETEXT,1 );
                inputname[5]=c;
                break;
        }
    }
}


/************************************************************************/
/* open_file(void)  : Opens all the necessary file handles, malloc()s   */
/*                    the space for the holding area, renames the saved */
/*                    game for backup, changes the two vital bytes at   */
/*                    decimal offset 322, then saves the new file.      */
/************************************************************************/

void open_files(void)
{
    unsigned int    culture_number=0;
    unsigned char   newname[128]="";
    struct   stat   statbuf;

    strncpy( newname, inputname,( strlen( inputname )-3) );
    strcat( newname, "BAK" );
    printf( "\nInputname=%s, newname=%s\n\n", inputname, newname );

    if( ( memarea = malloc (37856) ) == NULL )
        terminate_program( "Unable to allocate buffer of 37856 bytes for save file.\n",20 );

    if( ( CIVSVE_handle = open( inputname, O_RDONLY|O_BINARY)) < 1 )
        terminate_program( "Could not open the Civilization save file specified for reading.\n",10 );
    stat( inputname, &statbuf );
    if( statbuf.st_size != 37856 )
    {
        printf( "\nERROR! This file does not have a length corresponding to a Civilization\n" );
        printf( "save file. You might need a new version of CC (check with the author).\n" );
        terminate_program( "", 10 );
    }
    chk_read( CIVSVE_handle, memarea, 37856);
    _close( CIVSVE_handle );
    remove( newname );

    if( rename( inputname, newname ) != 0 )
        terminate_program( "Unable to rename Civilization file to backup file.\n",10);
    if( (CIVSVE_handle=open(inputname,O_CREAT|O_RDWR|O_BINARY,S_IREAD|S_IWRITE)) <1 )
        terminate_program("Could not open new Civilization save file.\n",10);

    culture_number = memarea[2]*2 + 312;
    memarea[culture_number] = 254;
    memarea[culture_number+1] = 127;

    chk_write( CIVSVE_handle, memarea, 37856);
    _close( CIVSVE_handle );

    printf( "\nOperation completed. File %s has been modified, original was copied\n",inputname );
    printf( "to backup filename %s. Any pre-existing version of %s\n",newname,newname );
    printf( "has been deleted.\n" );
}

/************************************************************************/
/* chk_read():  A "safe" _read function, checks for an incorrect read   */
/*              length automatically and errors out if an incomplete    */
/*              read occurs. Note: don't use for while() loops because  */
/*              this function will terminate HERE.                      */
/************************************************************************/

unsigned char chk_read( int f_handle, void *address, int length)
{
    if( _read( f_handle, address, length) != length )
    {
        printf("\nAn incomplete _read was performed.\n");
        if ( f_handle == CIVSVE_handle )
            terminate_program("Problem reading from specified Civilization save file\n",10);
        terminate_program("Problem reading from unknown file handle!\n",10);
    }
    return( 1 );
}

/************************************************************************/
/* chk_write(): A "safe" _write function, checks for an incorrect write */
/*              length automatically and errors out if an incomplete    */
/*              write occurs. Unlike chk_read this one is fine to use in*/
/*              loops because ideally none of these writes should ever  */
/*              error, if they do we need to halt immediately.          */
/************************************************************************/

unsigned char chk_write(int f_handle,void *address,int length)
{
    if(_write( f_handle, address, length) != length )
    {
        printf("\nAn incomplete _write was performed. It is likely the disk is full.\n");
        if ( f_handle == CIVSVE_handle )
            terminate_program("Problem writing to specified save file\n",10);
        terminate_program("Problem writing to unknown file handle (possibly CON:)\n",10);
    }
    return( 1 );
}


/********************************************************************/
/* parse_file_error():  Just checks the errno variable (see errno.h)*/
/*                      and responds with relavent string.          */
/********************************************************************/

void parse_file_error(void)
{
    printf("DOS ERROR returned: ");
    switch(errno)
    {
        case ENOENT:    printf("Path or file not found.\n");
                        break;
        case EMFILE:    printf("Too many open files.\n");
                        break;
        case EACCES:    printf("Permission denied.\n");
                        break;
        case EBADF:     printf("Bad file number.\n");
                        break;
        case E2BIG:     printf("Arg list to big on spawned operation.\n");
                        break;
        case EINVAL:    printf("Invalid argument for spawned process.\n");
                        break;
        case ENOEXEC:   printf("Exec format error on spawned process.\n");
                        break;
        case ENOMEM:    printf("Not enough memory for spawned process.\n");
                        break;
        default:    printf("Error creating/opening file.\n");
    }
}

/********************************************************************/
/* terminate_program(): Prints whatever message (may or may not be  */
/*                      be error condition) and passes error_value  */
/*                      to DOS command line.                        */
/********************************************************************/

void terminate_program(char *error_string,unsigned int error_value)
{
    if( ( error_value == 10 ) || ( error_value == 15 ) )
        parse_file_error();
    if( CIVSVE_handle )
        _close(CIVSVE_handle);
    if( memarea )
        free( memarea );
    printf( error_string );
    _exit( error_value );
}

/********************************************************************/
/* main():  Just calls the functions in order                       */
/********************************************************************/

int main(int argc,char *argv[])
{
    printf("\nCC (Civilization Cheat) 1.0 - grants you $30K in the bank\n");
    parse_command_line(argc,argv);
    open_files();
    terminate_program("\nProgram completed normally.\n",0);
    return(20);   /* This line never executed, program exists in line above */
}
